unit u_updownload;

{$mode delphi}

interface

uses
  Classes, SysUtils, Math, uempaquetar, uDataSetGenerico, urosx, uConstantesSimSEE;

const
  MAX_FILE_BLOCK = 1024*256;
// si la cantidad de bytes a almacenar supera MAX_FILE_BLOCK los datos
// se parten en bloques de MAX_FILE_BLOCK y se almacenan en una tabla de blocks
// aparte. Esto es así para evitar porblemas de transimisión de datos sobre http


// sube el archivo a la carpeta identificada por nid_carpeta.
// si no detecta error retorna el nid del archivo asignado en la DB.
// si hay error retorna -1.
function upload_archi(tbl_archivo: string; nid_carpeta: integer; archi: string): integer;

// Crea la carpeta "path" en la tabla tbl_carpeta
// devuelve nid_Carpeta o -1 en error
function crear_carpeta (tbl_carpeta, tbl_archivos, path: string; tipo: integer;
                       descripcion: string; vaciarSiExiste:Boolean=false):Integer;

// Elimina todos los archivos (y sus bloques) asociados a la carpeta
// nid_carpeta
procedure clear_carpeta(tbl_archivo: string; nid_carpeta: integer);


// Baja los archivo asociados a nidCarpeta y los almacena en la carpeta
// destino. Si no se pasa CarpetaDestino, o se pasa un string vacio se
// utiliza como carpeta destino el directorio de SimSEE local concatenado
// con el valor del campo "path" en el registro de la carpeta en la DB.
// Si el parámetro xflg se opmite o es verdadero, y si el nombre de archivo
// contiene el string: ".zip" se descomprime en una sub-carpeta
// con igual nombre de archivo y luego se borra el arhivo. Si xunzip= false
// no se intenta descomprimir el archivo.
function bajarCarpeta(tbl_carpeta, tbl_archivo: string; nidCarpeta: integer;
  carpetaDestino: string = ''; xunzip: boolean = True): boolean;

// en caso de error de alguna de las funciones anteriores, llamando
// esta función se puede obtener información sobre el error ocurrido.
function UltimoError: string;

implementation

var
  error: string;

function UltimoError: string;
begin
  Result := error;
end;

function bajarCarpeta(tbl_carpeta, tbl_archivo: string; nidCarpeta: integer;
  carpetaDestino: string = ''; xunzip: boolean = True): boolean;
var
  pathCarpeta, nomArch: string;
  chkSumRemoto, chkSumLocal: ShortString;
  orden: string;
  f: TextFile;
  ds, ds_Bloques: TResultadoQuery;
  rCarpeta, rArchi, rBloque: TDataRecord;
  datos: string;
  sql: string;
  nbloques: integer;
  k_bloque: integer;
  nidArchivo: integer;

begin
  Result := False;
  sql := 'SELECT fecha, path, descripcion FROM ' + tbl_carpeta +
    ' WHERE nid=' + IntToStr(nidCarpeta) + ' LIMIT 1';
  rCarpeta := sql_ficha(sql);

  if rCarpeta = nil then
  begin
    error := 'No se encontró el Rcord de la carpeta: ' + sql;
    Result := False;
    exit;
  end;
  if carpetaDestino = '' then
    pathCarpeta := getDir_SimSEE + rCarpeta.GetByNameAsString('path')
  else
    pathCarpeta := carpetaDestino;

  orden := 'SELECT nid, nombre, checksum, datos, nbloques FROM ' +
    tbl_archivo + ' WHERE nid_carpeta = ' + IntToStr(nidCarpeta) + ' LIMIT 1 ';
  ds := sql_query(orden);

  if ds = nil then
    raise Exception.Create(urosx.UltimoError);

  ForceDirectories(pathCarpeta);

  error := '';

  rArchi := ds.First;
  while not ds.EOF do
  begin
    nidArchivo := rArchi.GetByNameAsInt('nid');
    nomArch := rArchi.GetByNameAsString('nombre');
    nbloques := rArchi.GetByNameAsInt('nbloques');
    if nbloques = 1 then
    begin
      // baja archivo monobloque
      chkSumRemoto := rArchi.GetByNameAsString('checksum');
      datos := rArchi.GetByNameAsString('datos');
      chkSumLocal := checksum(@datos[1], length(datos));
      if chkSumLocal = chkSumRemoto then
      begin
        AssignFile(f, pathCarpeta + BarraDir + nomArch);
        Rewrite(f);
        Write(f, datos);
        CloseFile(f);

        if xunzip and (pos('.zip', nomArch) > 0) then
        begin
          uempaquetar.Unzip(pathCarpeta + BarraDir + nomArch);
          deletefile(pathCarpeta + BarraDir + nomArch);
        end;
      end
      else
      begin
        error := error + nomArch + #13#10;
      end;
    end
    else
    begin
      // baja archivo multibloque
      AssignFile(f, pathCarpeta + BarraDir + nomArch);
      Rewrite(f);

      orden := 'SELECT checksum, datos FROM ' + tbl_archivo +
        '_bloques WHERE nid_archivo = ' + IntToStr(nidArchivo) + ' ORDER BY k_bloque ';
      ds_Bloques := sql_query(orden);

      rBloque := ds_Bloques.First;

      while not ds_Bloques.EOF do
      begin
        chkSumRemoto := rBloque.GetByNameAsString('checksum');
        datos := rBloque.GetByNameAsString('datos');
        chkSumLocal := checksum(@datos[1], length(datos));
        if chkSumLocal = chkSumRemoto then
        begin
          Write(f, datos);
        end
        else
        begin
          error := error + nomArch + #13#10;
        end;
        rBloque := ds_bloques.Next;
      end;
      CloseFile(f);
      if xunzip and (pos('.zip', nomArch) > 0) then
      begin
        uempaquetar.Unzip(pathCarpeta + BarraDir + nomArch);
        deletefile(pathCarpeta + BarraDir + nomArch);
      end;
      ds_Bloques.Free;
    end;
    rArchi := ds.Next;
  end;
  ds.Free;
  Result := error = '';
end;

function crear_carpeta(tbl_carpeta, tbl_archivos, path: string; tipo: integer;
  descripcion: string; vaciarSiExiste: Boolean): Integer;
var
  nid_carpeta: integer;
  cnt: integer;
begin
  Result := -1;
  cnt := StrToInt(urosx.sql_func(
    'SELECT count(*) FROM '+tbl_carpeta+' WHERE path=''' + path + ''' LIMIT 1'));
  if cnt = 0 then
  begin
    nid_carpeta:= sql_nextnid(tbl_carpeta);
    if sql_exec('INSERT INTO '+ tbl_carpeta+' ( nid, fecha, tipo, path, descripcion ) VALUES ('
       + IntToStr(nid_carpeta) + ', now(), ' +IntToStr(
       tipo) + ', ''' + path + ''', ''' + Descripcion + ''' ) ') then

       Result := nid_carpeta;
  end
  else
  begin
    nid_carpeta:= StrToInt (sql_func('SELECT nid FROM '+tbl_carpeta+' WHERE path=''' + path + ''' LIMIT 1'));
    if vaciarSiExiste then
      clear_carpeta(tbl_archivos, nid_carpeta);

  end;
  Result := nid_carpeta;


end;

procedure clear_carpeta(tbl_archivo: string; nid_carpeta: integer);
begin


  sql_exec('DELETE FROM ' + tbl_archivo + '_bloques USING ' + tbl_archivo
   +' LEFT JOIN '+ tbl_archivo+'_bloques ON '
   + tbl_archivo + '.nid = ' + tbl_archivo + '_bloques.nid_archivo '
   +' WHERE ' + tbl_archivo + '.nid_carpeta = ' + IntToStr(nid_carpeta)
    );

  sql_exec('DELETE FROM ' + tbl_archivo + ' WHERE nid_carpeta = ' +
    IntToStr(nid_carpeta));
end;

function upload_archi(tbl_archivo: string; nid_carpeta: integer; archi: string): integer;
var
  sql, fechaServidor, pathCompleto: string;
  nombreArchivo: string;
  arch: string;
  chksum: ShortString;
  nidArchivo, i: integer;
  ok: boolean;
  fb: file of byte;
  nbytes: longint;
  nidBloque: integer;
  k_bloque: integer;
  bytes_restantes: integer;
  flg_error: boolean;
begin
  error := '';

  pathCompleto := ExtractFilePath(archi);
  nombreArchivo := ExtractFileName(archi);

  nidArchivo := sql_nextnid(tbl_archivo);
  assignfile(fb, archi);
  reset(fb);
  nbytes := system.filesize(fb);


  if (nbytes <= MAX_FILE_BLOCK) then
  begin
    setlength(arch, nbytes);
    blockread(fb, arch[1], nbytes);
    closefile(fb);

    chksum := checksum(@arch[1], Length(arch));
    sql := 'INSERT INTO ' + tbl_archivo +
      ' (nid, fecha, nombre, nid_carpeta, datos, checksum, nbloques) ' +
      'VALUES (' + IntToStr(nidArchivo) + ', now(), ''' + nombreArchivo +
      ''', ' + IntToStr(nid_Carpeta) + ', ''' + escapeChars(arch) +
      ''', ''' + escapeChars(chksum) + ''', 1)';

    if not sql_exec(sql) then
    begin
      error := sql + #13#10 + urosx.UltimoError;
      Result := -1;
    end
    else
      Result := nidArchivo;
  end
  else
  begin
    k_bloque := 0;
    bytes_restantes := nbytes;
    flg_error := False;
    while (bytes_restantes > 0) and not flg_error do
    begin
      Inc(k_bloque);

      if bytes_restantes > MAX_FILE_BLOCK then
        nbytes := MAX_FILE_BLOCK
      else
        nbytes := bytes_restantes;

      setlength(arch, nbytes);
      blockread(fb, arch[1], nbytes);
      chksum := checksum(@arch[1], Length(arch));
      nidBloque := sql_nextnid(tbl_archivo + '_bloques');

      sql := 'INSERT INTO ' + tbl_archivo +
        '_bloques (nid, nid_archivo, k_bloque, datos, checksum) ' +
        ' VALUES (' + IntToStr(nidBloque) + ', ' + IntToStr(nidArchivo) +
        ', ' + IntToStr(k_bloque) + ', ''' + escapeChars(arch) +
        ''', ''' + escapeChars(chksum) + ''')';
      if not sql_exec(sql) then
      begin
        error := sql + #13#10 + urosx.UltimoError;
        flg_error := True;
      end
      else
      begin
        bytes_restantes := bytes_restantes - nbytes;
      end;
    end;
    closefile(fb);
    if flg_error then
    begin
      sql := 'DELETE FROM ' + tbl_archivo + '_bloques WHERE nid_archivo= ' +
        IntToStr(nidArchivo);
      sql_exec(sql);
      Result := -1;
    end
    else
    begin
      sql := 'INSERT INTO ' + tbl_archivo +
        ' (nid, fecha, nombre, nid_carpeta, datos, checksum, nbloques) ' +
        'VALUES (' + IntToStr(nidArchivo) + ', now(), ''' + nombreArchivo +
        ''', ' + IntToStr(nid_Carpeta) + ', '''', '''', ' + IntToStr(k_bloque) + ')';
      sql_exec(sql);
      Result := nidArchivo;
    end;
  end;
end;

begin
  error := '';
end.
