unit ugraficador;

{$mode delphi}

interface

uses
  Classes, Math, types, SysUtils, Graphics, xmatdefs, autoesca, ulogoSimSEE;

type
  TTipoGrafico = (TG_DispersionXY, TG_AreasApiladas, TG_Lineas, TG_Barras);
  TTipoSerie = (TS_Serie, TS_SerieX, TS_SerieY);
  TTipoMarcador = (TM_Circulo, TM_Cuadro, TM_Rombo, TM_Cruz, TM_X);
  TkColorPaleta = 0 .. 14;


const
  TColoresPaleta: array[TkColorPaleta] of TColor = (
    clBlack
    , clMaroon
    , clGreen
    , clOlive
    , clNavy
    , clPurple
    , clTeal
    , clGray
    , clSilver
    , clRed
    , clLime
    , clYellow
    , clBlue
    , clFuchsia
    , clAqua);

  TNombreColores: array[TkColorPaleta] of string = (
    'Negro'
    , 'Marrón'
    , 'Verde'
    , 'Oliva'
    , 'Azul marino'
    , 'Púrpura'
    , 'Verde azulado'
    , 'Gris'
    , 'Plata'
    , 'Rojo'
    , 'Lima'
    , 'Amarillo'
    , 'Azul'
    , 'Fusia'
    , 'Agua');

  TNombreMarcador: array[TTipoMarcador] of string = (
    'Circulo', 'Cuadro', 'Rombo', 'Cruz', 'X');

type
  TGrafico = class;

  TEjeGrafico = class
    titulo: string;
    v_min_real, v_max_real: NReal; // valores extremos de las series.
    v_min_forzado, v_max_forzado: NReal; // valores fijados forzados.
    v_min_auto, v_max_auto: NReal; // valores calculados automáticamente.
    v_min_activo, v_max_activo: NReal;
    // valores activos, auto o forzado según corresponda

    flg_min_auto, flg_max_auto: boolean; // indica se usa el automático o el forzado,
    v0: NReal; // posición primera linea de la grilla
    dv: NReal; // ancho de la divición
    nDivs: integer; // cantida de divisiones de la grilla

    flg_mostrar_lineas: boolean;

    constructor Create(Titulo: string);
    // en min_val, max_val deben ingresar los extremos REALES
    // y son devueltos los aplicables
    procedure Calcular(var min_val, max_val: NReal);
    // Calcula los valores automáticos en base
    procedure Forzar_vmin(v_min: NReal);
    procedure Forzar_vmax(v_max: NReal);
    procedure Librar_vmin;
    procedure Librar_vmax;
    procedure Free;
  end;

  TSerie = class
    nombre: string;
    orden: integer;
    colorArea, colorLinea, colorFondoMarcador, colorLineaMarcador: TColor;
    tipoMarcador: TTipoMarcador;
    tipoGrafico: TTipoGrafico; // AreaApilada, XY, linea
    tipoSerie: TTipoSerie;
    valores: TDAOfNReal;
    serie_x: TSerie;
    eje: integer; // 0 izquierdo, 1 derecho
    formato_str: string;

    // al crear la serie se pasa xvalores y la serie se encarga de eliminar
    constructor Create(nombre: string; xvalores: TDAOfNReal; serie_x: TSerie;
      grafico: TGrafico; formato_str: string; flg_SerieX: boolean = False);
    function FormatoFecha: boolean;
    procedure Free;

  end;

  { TGrafico }

  TGrafico = class
    nombre: string;
    Titulo: string;
    eje_x, eje_y, eje_y2: TEjeGrafico;
    series_y: TList;
    series_x: TList;
    tipoGrafico: TTipoGrafico;
    kEnLibro: integer;

    constructor Create(nombre: string; tipoGrafico: TTipoGrafico);
    procedure Draw(c: TCanvas; ancho, alto: integer);
    procedure AddSerie(NuevaSerie: TSerie);
    function ProximoColor: TColor;
    function ProximoMarcador: TTipoMarcador;
    procedure SaveJPG(archi: string; ancho, alto: integer);
    procedure Free;

  private
    nid_ProximoColor: integer;
    nid_ProximoMarcador: integer;

    w0, h0, w1, h1: integer;
    mx, my, my2: NReal;
    min_X, max_X, min_Y, max_Y, min_Y2, max_Y2: NReal;

    tag_w, tag_h: integer;
    dh_TItuloGrafico: integer;
    dh_TItuloX: integer;
    dw_TItuloY2: integer;
    dw_TItuloY: integer;
    dw_EtiquetasSeries: integer;
    dh_Etiqueta: integer;
    tag_Etiqueta: integer;
    dw_EscalaY: integer;
    dw_EscalaY2: integer;
    dh_EscalaX: integer;

    RadioMarcador: integer;

    function x2w(x: NReal): integer;
    function y2h(y: NReal): integer;
    function y2h2(y: NReal): integer;


    procedure plotMarcador(
              c: TCanvas;
              x, y: integer;
              tipoMarcador: TTipoMarcador;
              colorFondoMarcador, colorLineaMarcador: integer;
              anchoLinea: integer;
              RadioMarcador: integer );

    procedure PlotDispersionXY_marcadores(c: TCanvas; lst_SeriesDispXY: TList);

    procedure PlotDispersionXY( c: TCanvas; lst_SeriesDispXY: TList);
    procedure PlotAreasApiladas(c: TCanvas; lst_SeriesAreasApiladas: TList );

  end;

var
  clDefault: TColor;

implementation

constructor TEjeGrafico.Create(Titulo: string);
begin
  inherited Create;
  self.titulo := Titulo;
  v_min_real := 0;
  v_max_real := 0;
  v_min_forzado := 0;
  v_max_forzado := 0;
  v_min_activo := 0;
  v_max_activo := 0;
  dv := 1.0;
  nDivs := 10;
  v_min_auto := 0;
  v_max_auto := 0;
  flg_min_auto := True;
  flg_max_auto := True;
  flg_mostrar_lineas := True;
end;

procedure TEjeGrafico.Calcular(var min_val, max_val: NReal);
// Calcula los valores automáticos en base
begin
  self.v_min_real := min_val;
  self.v_max_real := max_val;

  v_min_auto := min_val;
  v_max_auto := max_val;

  if (flg_min_auto and flg_max_auto) then
    Escala125N(v_min_auto, v_max_auto, dv, nDivs, 1)
  else
    Escala125N(v_min_auto, v_max_auto, dv, nDivs, 0);

  v0 := v_min_auto;

  if flg_min_auto then
    v_min_activo := v_min_auto
  else
    v_min_activo := v_min_forzado;

  if flg_max_auto then
    v_max_activo := v_max_auto
  else
    v_max_activo := v_max_forzado;

  min_val := v_min_activo;
  max_val := v_max_activo;
end;


procedure TEjeGrafico.Forzar_vmin(v_min: NReal);
begin
  v_min_forzado := v_min;
  flg_min_auto := False;
end;

procedure TEjeGrafico.Forzar_vmax(v_max: NReal);
begin
  v_max_forzado := v_max;
  flg_max_auto := False;
end;

procedure TEjeGrafico.Librar_vmin;
begin
  flg_min_auto := True;
end;

procedure TEjeGrafico.Librar_vmax;
begin
  flg_max_auto := True;
end;


procedure TEjeGrafico.Free;
begin
  inherited Free;
end;


(* Métodos de TSerie *)
constructor TSerie.Create(nombre: string; xvalores: TDAOfNReal;
  serie_x: TSerie; grafico: TGrafico; formato_str: string; flg_SerieX: boolean = False);
begin
  inherited Create;
  self.Nombre := nombre;
  valores := xvalores;
  Self.Serie_x := serie_x;
  self.formato_str := formato_str;

  if grafico <> nil then
  begin
    colorArea := grafico.ProximoColor;
    colorLinea := colorArea;
    colorFondoMarcador := colorArea;
    colorLineaMarcador := colorArea;
    tipoMarcador := grafico.ProximoMarcador;
    tipoGrafico := grafico.tipoGrafico;
  end
  else
  begin
    colorArea := clBlack;
    colorLinea := clBlack;
    colorFondoMarcador := clBlack;
    colorLineaMarcador := clBlack;
    tipoMarcador := TM_Cruz;
    tipoGrafico := TG_Lineas;
  end;

  if flg_SerieX then
    tipoSerie := TS_SerieX
  else if serie_x = nil then
    tipoSerie := TS_Serie
  else
    tipoSerie := TS_SerieY;

  eje := 0; // por defecto del lado izquiedo.
end;

function TSerie.FormatoFecha: boolean;
begin
  Result := (pos('-', formato_str) + pos('/', formato_str)) > 0;
end;

procedure TSerie.Free;
begin
  setlength(valores, 0);
  inherited Free;
end;

(* Métodos de TGrafico *)
constructor TGrafico.Create(nombre: string; tipoGrafico: TTipoGrafico);
begin
  inherited Create;
  self.Nombre := nombre;
  titulo := '';
  eje_x := TEjeGrafico.Create('');
  eje_y := TEjeGrafico.Create('');
  eje_y2 := TEjeGrafico.Create('');
  series_y := TList.Create;
  series_x := TList.Create;
  tipoGrafico := TG_Lineas;
  nid_ProximoColor := 0;
  nid_ProximoMarcador := 0;
  kEnLibro:= -1;
end;



procedure TGrafico.AddSerie(NuevaSerie: TSerie);
begin
  if NuevaSerie.tipoSerie = TS_SerieX then
    series_x.add(NuevaSerie)
  else
    series_y.add(NuevaSerie);
end;

procedure AmpliarRango(var amin, amax: NReal; var cnt_series: integer;
  NuevaSerie: TSerie);
var
  bmin, bmax: NReal;
begin
  bmin := vmin(NuevaSerie.valores);
  bmax := vmax(NuevaSerie.Valores);
  if cnt_series > 0 then
  begin
    if bmin < amin then
      amin := bmin;
    if bmax > amax then
      amax := bmax;
  end
  else
  begin
    amin := bmin;
    amax := bmax;
  end;
  Inc(cnt_series);
end;


function TGrafico.x2w(x: NReal): integer;
begin
  Result := w0 + trunc(mx * (x - min_X) + 0.5);
end;

function TGrafico.y2h(y: NReal): integer;
begin
  Result := h1 - trunc(my * (y - min_Y) + 0.5);
end;

function TGrafico.y2h2(y: NReal): integer;
begin
  Result := h1 - trunc(my2 * (y - min_Y2) + 0.5);
end;


procedure TGrafico.plotMarcador(
          c: TCanvas;
          x, y: integer;
          tipoMarcador: TTipoMarcador;
          colorFondoMarcador, colorLineaMarcador: integer;
          anchoLinea: integer;
          RadioMarcador: integer );

var
  rombo: array of TPoint;
  x1, y1, x2, y2: integer;
begin
  x1:= x - RadioMarcador;
  x2:= x + RadioMarcador;
  y1:= y - RadioMarcador;
  y2:= y + RadioMarcador;
  case tipoMarcador of
    TM_Circulo:
       c.Ellipse( x1, y1, x2, y2 );
    TM_Cuadro:
       c.Rectangle(x1, y1, x2, y2 );
    TM_Rombo:
       begin
         setlength( rombo, 4 );
         rombo[0].x:= x; rombo[0].y:= y - RadioMarcador;
         rombo[1].x:= x - RadioMarcador; rombo[1].y:= y;
         rombo[2].x:= x; rombo[2].y:= y + RadioMarcador;
         rombo[3].x:= x + RadioMarcador; rombo[3].y:= y;
         c.Polygon( rombo );
         setlength( rombo, 0 );
       end;
    TM_Cruz:
       begin
         c.Line( x1, y, x2, y);
         c.Line( x, y1, x, y2);
       end;

    TM_X:
       begin
         c.Line( x1, y1, x2, y2);
         c.Line( x1, y2, x2, y1);
       end;
  end;
end;




procedure TGrafico.PlotDispersionXY_marcadores(c: TCanvas; lst_SeriesDispXY: TList);
var
  k, j: integer;
  eje: integer;
  aSerie, aSerieX: TSerie;
  x, y: integer;
begin
  aSerie := lst_SeriesDispXY[0];
  eje:= aSerie.eje;
  for k := 0 to lst_SeriesDispXY.Count - 1 do
  begin
    aSerie := lst_SeriesDispXY[k];
    aSerieX := aSerie.serie_x;
    c.Pen.Color := aSerie.colorLinea;
    c.Pen.Width:= 3;
    c.Brush.Color:= aSerie.colorFondoMarcador;
    c.Brush.Style:= bsClear;

    for j := 0 to high(aSerie.valores) do
    begin
      x := x2w(aSerieX.valores[j]);
      if aSerie.eje = 0 then
        y := y2h(aSerie.valores[j])
      else
        y := y2h2(aSerie.valores[j]);
      plotMarcador(c, x, y, aSerie.tipoMarcador, aSerie.colorFondoMarcador, aSerie.colorLineaMarcador, 3, RadioMarcador );
    end;
  end;
end;


procedure TGrafico.PlotDispersionXY(c: TCanvas; lst_SeriesDispXY: TList);
var
  k, j: integer;
  eje: integer;
  aSerie, aSerieX: TSerie;
  x, y: integer;
begin
  aSerie := lst_SeriesDispXY[0];
  eje:= aSerie.eje;
  for k := 0 to lst_SeriesDispXY.Count - 1 do
  begin
    aSerie := lst_SeriesDispXY[k];
    aSerieX := aSerie.serie_x;
    c.Pen.Color := aSerie.colorLinea;
    c.Pen.Width:= 3;

    x := x2w(aSerieX.valores[0]);
    if eje = 0 then
      y := y2h(aSerie.valores[0])
    else
      y := y2h2(aSerie.valores[0]);
    c.MoveTo(x, y);

    for j := 1 to high(aSerie.valores) do
    begin
      x := x2w(aSerieX.valores[j]);
      if aSerie.eje = 0 then
        y := y2h(aSerie.valores[j])
      else
        y := y2h2(aSerie.valores[j]);
      c.LineTo(x, y);
    end;
  end;
end;

procedure TGrafico.PlotAreasApiladas(c: TCanvas; lst_SeriesAreasApiladas: TList);
var
  kSerie: integer;
  k: integer;
  vinf, vsup: TDAOfNreal;
  poligono: array of TPoint;
  aSerieX, aSerie: TSerie;
  vx, vy: NReal;
  NP: integer;
  NPM1: integer;
  eje: integer;

begin
  aSerie := lst_SeriesAreasApiladas[0];
  eje:= aSerie.eje;
  aSerieX := aSerie.serie_x;
  NP:=length(aSerie.valores);
  NPM1:= NP-1;

  setlength(vinf, NP);
  setlength(vsup, NP);
  setlength(poligono, 2 * NP );

  vclear(vinf);
  for k := 0 to NPM1 do
    vsup[k] := aSerie.valores[k];

  kSerie := 0;
  repeat
    c.Brush.color := aSerie.colorArea;

    for k := 0 to NPM1 do
    begin
      vx := aSerieX.valores[k];
      vy := vsup[k];
      poligono[k].x := x2w(vx);
      if eje = 0 then
        poligono[k].y := y2h(vy)
      else
        poligono[k].y := y2h2(vy);
    end;
    for k := 0 to NPM1 do
    begin
      vx := aSerieX.valores[NPM1-k];
      vy := vinf[NPM1-k];
      poligono[k + NP].x := x2w(vx);
      if eje = 0 then
         poligono[k + NP].y := y2h(vy)
      else
          poligono[k + NP].y := y2h(vy);

    end;


    c.Polygon(poligono, true);

    Inc(kSerie);
    if kSerie < lst_SeriesAreasApiladas.Count then
    begin
      aSerie := lst_SeriesAreasApiladas[kSerie];
      vswap( vsup, vinf );

      for k := 0 to NPM1 do
        vsup[k] := vinf[k] + aSerie.valores[k];
    end;

  until kSerie >= lst_SeriesAreasApiladas.Count;

  setlength( vsup, 0 );
  setlength( vinf, 0 );
  setlength( poligono, 0 );

end;


procedure AmpliarRangoApiladas( var min_Y, max_Y: NReal; var cnt_seriesConsideradas: integer; lst_SeriesAA: TList );
var
  vsup: TDAOfNReal;
  aSerie: TSerie;
  k, j: integer;
  N: integer;
begin
  aSerie:= lst_SeriesAA[0];
  N:= length( aSerie.valores );

  setlength( vsup, N );
  vclear( vsup );

  if cnt_seriesConsideradas = 0 then
  begin
    min_Y:= 0;
    max_Y:= 0;
  end;
  inc( cnt_seriesConsideradas );

  for k:= 0 to lst_SeriesAA.count -1 do
  begin
     aSerie:= lst_SeriesAA[k];
     for j:= 0 to high( vsup ) do
     begin
         vsup[j]:= vsup[j] +  aSerie.valores[j];
         if vsup[j] < min_Y then
            min_Y := vsup[j]
         else if vsup[j] > max_Y then
            max_Y := vsup[j];
     end;
  end;

end;


procedure TGrafico.Draw(c: TCanvas; ancho, alto: integer);
var
  w, h: integer;
  x, y: integer;
  x_v, y_v: NReal;

  NSeries_X: integer;
  NSeries_Y: integer;
  NSeries_XY: integer;
  amin, amax: NReal;
  k, j: integer;
  aSerieX, aSerie: TSerie;
  cnt_series_rango: integer;
  cnt_series_rango2: integer;
  LargoSeries_noxy: integer; // cantidad de elementos de la serie más larga que no es xy

  // Listas para ordenar el dibujo
  lst_barras_1, lst_barras_2, lst_lineas_1, lst_lineas_2, lst_dispXY_1,
  lst_dispXY_2, lst_AApiladas_1, lst_AApiladas_2: TList;

  flg_X_es_fecha: boolean;

  s: string;
  dw_s, dh_s: integer;

begin
  lst_barras_1 := TList.Create;
  lst_barras_2 := TList.Create;
  lst_lineas_1 := TList.Create;
  lst_lineas_2 := TList.Create;
  lst_dispXY_1 := TList.Create;
  lst_dispXY_2 := TList.Create;
  lst_AApiladas_1 := TList.Create;
  lst_AApiladas_2 := TList.Create;

  NSeries_XY := 0;
  largoSeries_noxy := 0;
  flg_X_es_fecha := False;

  NSeries_X := series_x.Count;
  if NSeries_X > 0 then
  begin
    cnt_series_rango := 0;
    for k := 0 to NSeries_X - 1 do
    begin
      AmpliarRango(min_X, max_X, cnt_series_rango, series_X[k]);
      flg_X_es_fecha := flg_X_es_fecha or TSerie(series_X[k]).FormatoFecha;
    end;
  end;

  cnt_series_rango := 0;
  cnt_series_rango2 := 0;

  NSeries_Y := series_y.Count;
  if NSeries_Y > 0 then
  begin
    for k := 0 to NSeries_Y - 1 do
    begin
      aSerie := series_y[k];

      if aSerie.serie_x <> nil then
        Inc(NSeries_XY)
      else
      begin
        if length(aSerie.valores) > LargoSeries_noxy then
          LargoSeries_noxy := length(aSerie.valores);
      end;

//  writeln( 'Serie: ', aSerie.nombre,', TipoGrafico_: ', Ord( aSerie.TipoGrafico ),', eje: ', aSerie.eje, ', Marcador: ', IntToStr( ord( aSerie.tipoMarcador )) );

      if aSerie.eje = 0 then
      begin
        if aSerie.tipoGrafico <> TG_AreasApiladas then
          AmpliarRango(min_Y, max_Y, cnt_series_rango, aSerie);
        case aSerie.tipoGrafico of
          TG_Lineas: lst_lineas_1.add(aSerie);
          TG_AreasApiladas: lst_AApiladas_1.add(aSerie);
          TG_Barras: lst_barras_1.add(aSerie);
          TG_DispersionXY: lst_dispXY_1.add(aSerie);
        end;
      end
      else
      begin
        if aSerie.tipoGrafico <> TG_AreasApiladas then
            AmpliarRango(min_Y2, max_Y2, cnt_series_rango2, aSerie);
        case aSerie.tipoGrafico of
          TG_Lineas: lst_lineas_2.add(aSerie);
          TG_AreasApiladas: lst_AApiladas_2.add(aSerie);
          TG_Barras: lst_barras_2.add(aSerie);
          TG_DispersionXY: lst_dispXY_2.add(aSerie);
        end;
      end;
    end;
  end;


  // ahora si hay areas apiladas recalculamos los rangos
  if lst_AApiladas_1.count > 0 then
     AmpliarRangoApiladas( min_Y, max_Y, cnt_series_rango, lst_AApiladas_1 );

  if lst_AApiladas_2.count > 0 then
     AmpliarRangoApiladas( min_Y2, max_Y2, cnt_series_rango2, lst_AApiladas_2 );


  c.Font.Orientation:= 0;
  c.Font.Size := 14;

  RadioMarcador:= 4;
  tag_w:= 8;
  tag_h:= 8;
  if self.Titulo <> '' then
     dh_TItuloGrafico:= c.TextHeight( 'M' )
  else
     dh_TituloGrafico:= 0;

  if self.eje_x.titulo <> '' then
     dh_TItuloX:= c.TextHeight( 'M' )
  else
     dh_TituloX:= 0;

  c.Font.Orientation:= 900;

  if self.eje_y2.titulo <> '' then
    dw_TItuloY2:= c.TextHeight( 'M' )
  else
    dw_TItuloY2:=0;

  if self.eje_y.titulo <> '' then
     dw_TItuloY:= c.TextHeight( 'M' )
  else
     dw_TItuloY:= 0;

  dh_EscalaX:= c.TextWidth( '12/06/2015 12:00' );
  c.Font.Orientation:= 0;

  dw_EtiquetasSeries:= 0;
  for k := 0 to NSeries_Y - 1 do
  begin
    aSerie := series_y[k];
    dw_EtiquetasSeries := max(dw_EtiquetasSeries, c.TextWidth(aSerie.nombre));
  end;
  dh_Etiqueta:= c.TextHeight( 'M' );
  tag_Etiqueta:= 2;


  if cnt_series_rango > 0 then
    dw_EscalaY:= c.TextWidth( '#####.##' )
  else
    dw_EscalaY:= 0;

  if cnt_series_rango2 > 0 then
    dw_EscalaY2:= c.TextWidth( '#####.##' )
  else
    dw_EscalaY2:= 0;

  w := ancho;
  h := alto;



  w0 := tag_w+ dw_TItuloY + tag_w + dw_EscalaY+ tag_w; // borde izquirdo a calcualr
  w1 := w - tag_w - dw_EtiquetasSeries - tag_w - dw_EscalaY2 - tag_w -RadioMarcador - tag_w ; // borde derecho
  h0 := tag_h+ dh_TItuloGrafico + tag_h; // borde superior
  h1 := h - tag_h - dh_TItuloX - tag_h - dh_EscalaX; // borde inf


  if flg_X_es_fecha then
  begin
      eje_x.Forzar_vmin( min_x );
      eje_x.Forzar_vmax( max_x );
  end;

  eje_x.Calcular(min_x, max_x);
  eje_y.Calcular(min_y, max_y);
  eje_y2.Calcular(min_y2, max_y2);


  mx := (w1 - w0) / (max_x - min_x);
  if cnt_series_rango > 0 then
    my := (h1 - h0) / (max_y - min_y);

  if cnt_series_rango2 > 0 then
    my2 := (h1 - h0) / (max_y2 - min_y2);

  // borramos
  c.Brush.Color := clWhite;
  c.FillRect(0, 0, w - 1, h - 1);
  c.Pen.Style:= psSolid;
  c.Pen.Color := clBlack;
  c.Rectangle(w0, h0, w1, h1);

  c.TextOut(trunc( (w1 + w0) / 2.0 - c.TextWidth(Titulo) / 2.0),
    tag_h, AnsiToUtf8( Titulo ) );

  x := w - tag_w - dw_EtiquetasSeries ;
  y := h0 + tag_h;

  for k := 0 to NSeries_Y - 1 do
  begin
    aSerie := series_y[k];
    c.Pen.Color := aSerie.colorLinea;
    c.Pen.Width:= 3;
    c.Brush.Color:= aSerie.colorFondoMarcador;
    c.Brush.Style:= bsClear;

    plotMarcador( c, x- RadioMarcador - tag_w , trunc( y + k * (dh_Etiqueta + tag_Etiqueta) + dh_Etiqueta/2 + 0.5),
      aSerie.tipoMarcador,   aSerie.colorFondoMarcador, aSerie.colorLineaMarcador, 3, RadioMarcador );

    c.Pen.Color := clBlack;
    c.TextOut(x, y + k * (dh_Etiqueta + tag_Etiqueta), aSerie.nombre);
  end;


  c.ClipRect := Rect(w0, h0, w1, h1);
  c.Clipping := True;

  // Plot AreasApiladas
  c.Pen.Width:= 1;
  if lst_AApiladas_1.count > 0 then
     PlotAreasApiladas(c, lst_AApiladas_1 );
  if lst_AApiladas_2.count > 0 then
       PlotAreasApiladas(c, lst_AApiladas_2);


  c.Pen.Width:= 3;
    // Plot DispersiónXY
    if lst_dispXY_1.count > 0 then
    begin
      PlotDispersionXY( c, lst_dispXY_1 );
      PlotDispersionXY_marcadores( c, lst_dispXY_1 );
    end;
    if lst_dispXY_2.count > 0 then
    begin
      PlotDispersionXY( c, lst_dispXY_2 );
      PlotDispersionXY_marcadores( c, lst_dispXY_2 );
    end;

  // GRILLAS
  c.ClipRect := Rect(0, 0, w, h);
  c.Clipping := True;
  c.Pen.Width:= 1;
  c.Pen.Color := clBlack;
  c.Pen.Style := psSolid;
  c.Brush.Color:= clWhite;
  c.Brush.Style:= bsSolid;


  c.Font.Orientation:= 900;
  // Etiquetas X
  dw_s:= c.TextHeight( '1' ) div 2;
  for k := 0 to eje_x.nDivs do
  begin
    x_v := Eje_x.v0 + k * eje_x.dv;
    x := x2w(x_v);

    if flg_X_es_fecha then
       s:= DateTimeToStr(x_v)
    else
      s:= FloatToStrF(x_v, ffGeneral, 12, 2);
    c.TextOut(x -dw_s, h1 + tag_h + c.TextWidth( s ), s)
  end;
  c.Font.Orientation:= 0;

  // Etiquetas Y
  dh_s:= c.TextHeight( '1' ) div 2;
  for k := 0 to eje_y.nDivs do
  begin
    y_v := Eje_y.v0 + k * eje_y.dv;
    y := y2h(y_v);
    s:= FloatToStrF(y_v, ffGeneral, 12, 2);
    c.TextOut(w0 - tag_w - c.TextWidth(s), y - dh_s, s);
  end;


  // Etiquetas Y2
  if cnt_series_rango2 > 0 then
  begin
    dh_s:= c.TextHeight( '1' ) div 2;
    for k := 0 to eje_y2.nDivs do
    begin
      y_v := Eje_y2.v0 + k * eje_y2.dv;
      y := y2h2(y_v);
      s:= FloatToStrF(y_v, ffGeneral, 12, 2);
      c.TextOut(w1 + tag_w , y - dh_s, s);
    end;

    c.Pen.Color:= clRed;
    c.Pen.Style := psDot;
    for k := 0 to eje_y2.nDivs do
    begin
      y_v := Eje_y2.v0 + k * eje_y2.dv;
      y := y2h2(y_v);
      c.Line( w1 - 20, y, w1, y );
    end;
  end;



  c.Pen.Color:= clBlack;
  c.Font.Orientation := 900;
  y := c.TextHeight(eje_y.titulo);
  c.TextOut( tag_w, trunc((h - y) / 2), AnsiToUtf8( eje_y.titulo));

  c.Font.Orientation := 0;
  c.Pen.Color := clBlue;
  c.Pen.Style := psDot;

  // GridX
  if eje_X.flg_mostrar_lineas then
    for k := 0 to eje_x.nDivs  do
    begin
      x_v := Eje_x.v0 + k * eje_x.dv;
      x := x2w(x_v);
      c.MoveTo(x, h1 );
      c.LineTo(x, h0 );
    end;


  // Grid_y
  if eje_X.flg_mostrar_lineas then
    for k := 1 to eje_y.nDivs - 1 do
    begin
      y_v := Eje_y.v0 + k * eje_y.dv;
      y := y2h(y_v);
      c.MoveTo(w0 , y);
      c.LineTo(w1 , y);
    end;

  ulogoSimSEE.DataModule1.img_logoSimSEE.Draw( c, w1 - 140, h0 - 16, 0 );


end;

function TGrafico.ProximoColor: TColor;
begin
  Result := TColoresPaleta[TkColorPaleta(nid_ProximoColor)];
  nid_ProximoColor := (nid_ProximoColor + 1) mod ( High(TkColorPaleta) + 1);
end;

function TGrafico.ProximoMarcador: TTipoMarcador;
begin
  Result := TTipoMarcador(nid_ProximoMarcador);
  nid_ProximoMarcador := (nid_ProximoMarcador + 1) mod ( Ord(High(TTipoMarcador)) + 1 );
end;




procedure TGrafico.SaveJPG(archi: string; ancho, alto: integer);
var
  jp: TJPEGImage;  //Requires the "jpeg" unit added to "uses" clause.
  b: TBitmap;
begin
  b := TBitmap.Create;
  b.Width := ancho;
  b.Height := alto;
  Draw(b.Canvas, ancho, alto);
  jp := TJPEGImage.Create;
  try
    jp.Assign(b);
    jp.SaveToFile(archi);
  finally
    jp.Free;
  end;
  b.Free;
end;


procedure TGrafico.Free;
var
  k: integer;
  aSerie: TSerie;
begin

  for k := 0 to series_y.Count do
  begin
    aSerie := series_y[k];
    aSerie.Free;
  end;

  for k := 0 to series_x.Count do
  begin
    aSerie := series_x[k];
    aSerie.Free;
  end;
  eje_x.Free;
  eje_y.Free;
  eje_y2.Free;
  inherited Free;

end;


begin
  clDefault:= StringToColor( 'clDefault' )
end.
